from __future__ import print_function
from __future__ import absolute_import
#
# Global Cache network devices - functional devices
#
# Copyright 2010-2016,2019 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import deque
import re
import sys
import socket
import struct
import time
import math

import asyn
import asyn.resolve
import asyn.utility

from . import gcir
from .core import Error
from .core import DEVICE_GC100, DEVICE_ITACH, DEVICE_FLEX, UDP_PORT, UDP_ADDRESS
from .core import PULSE_MIN, PULSE_MIN_FLEX


#
# An IR transmission job
#
class IRTask(object):
	""" Describe a single (queued or ongoing) IR transmission. """
	def __init__(self, device, ir, repeat):
		self.dev = device						# address of transmitting device
		self.ir = ir							# code being sent
		self.repeat = repeat					# number of repeats desired
		self.start = None						# not yet sent
		self.seq = None							# no sequence yet

	def transmit(self):
		assert(self.start is None)
		self.seq = self.dev._transmit(self.ir, self.repeat)
		self.start = time.time()

	def __repr__(self):
		return "<IRTask#%s %s<-%s{%s}>" % (self.seq, self.dev, self.ir, self.repeat)


#
# Common features of all GC functional devices
#
class Device(asyn.Callable):
	""" Some hardware at a particular connector address of a GCNet.

		By definition, this performs one particular function (IR, SENSOR, SERIAL, etc.)
		We don't support dynamic reconfiguration here, so this is a constant function,
		and we directly represent it as methods on subclasses.
	"""
	def __init__(self, netdev, address, specific):
		asyn.Callable.__init__(self)
		self.hostdev = netdev
		self.address = address
		self.specific = specific

	def close(self):
		pass

	@property
	def name(self):
		return "%s@%s" % (self.hostdev.hostname, self.address)

	def command(self, cmd, *args):
		""" Issue a command directly to this particular device. """
		self.hostdev.command(','.join([cmd, self.address] + list(map(str, args))))

	def _calldown(self, ctx, cmd, args):
		""" Default callout - just complain. """
		self.callout(Error("unexpected event", self.type, cmd, self.address, args))

	def __repr__(self):
		return "<%s %s/%s>" % (self.type, self.address, self.specific)


#
# IR Emitter devices
#
class IREmitter(Device):
	""" An IR emitter device. Can send IR codes.

		Call emit(symbol, count) to transmit an IR symbol. Transmissions
		can be queued to arbitrary length, and symbols will be sent back-to-back
		until the queue is empty.
		Calls out an IRTask object when its transmission is complete,
		and calls out an additional None whenever the queue becomes empty.
		Note that on some GC devices, all IREmitters share a single queue,
		while others have a queue per emitter. Only completions on this device
		are reported, but queue-empty (None) may be shared or not.
	"""
	type = "IR"

	def __init__(self, netdev, address, specific):
		Device.__init__(self, netdev, address, specific)
		self._ir = None
		if netdev._irqueue is not None:
			self._irqueue = netdev._irqueue				# share netdev queue
		else:
			self._irqueue = deque()						# run our own queue

	def emit(self, ir, repeat=0):
		""" Queue ir for emission, with given repeat count. """
		task = IRTask(self, ir.symbol().code(repeat), repeat)
		self._irqueue.append(task)
		if self._irqueue[0] == task:					# was empty
			task.transmit()								# start transmission

	def queue(self):
		return self._irqueue

	def _transmit(self, ir, repeat):
		""" Issue GC-specific transmission command. Returns sequence number. """
		self.hostdev._irseq += 1
		seq = self.hostdev._irseq
		count = 1
		offset = 1
		(base, repeat, repeat_count) = ir.form()
		pulses = base
		if repeat_count > 0:
			count += repeat_count
			if repeat and repeat != base:
				pulses = pulses + repeat
				offset += 2 * len(base)
		
		# impose PLUSE_MIN seconds minimum on each pulse leg (iTach requirement)
		pulse_min = int(math.ceil(PULSE_MIN * ir.carrier))
		if self.hostdev.type == DEVICE_FLEX:			# iTach Flex doesn't obey its own rules...
			pulse_min = max(pulse_min, PULSE_MIN_FLEX)	# experimental floor
		def xlate(pulse):
			return pulse.form(pulse_min)
		
		self.command("sendir",
			"%d,%d,%d,%d,%s" % (seq, ir.carrier, count, offset, ','.join(map(xlate, pulses))))
		return seq

	def _calldown(self, ctx, cmd, args):
		""" Catch transmission-complete message. """
		if cmd == "completeir":
			self._ircomplete(args)
		else:
			Device._calldown(self, ctx, cmd, args)

	def _ircomplete(self, seq):
		""" Called when the transmitter reports a completed transmission. """
		old_task = self._irqueue.popleft()				# pop off transmit head
		assert int(seq) == old_task.seq
		self.callout('transmitted', old_task)			# notify upstream
		if self._irqueue:								# if queue is not empty...
			self._irqueue[0].transmit()					# ... push next element
		else:
			self.callout('empty')						# else notify upstream the queue is empty


#
# LED light dimmer devices.
# Placeholder - no hardware or desire.
#
class LEDLight(Device):
	""" A LED PWM dimmer. Can drive the dimming control of certain LED light controllers.

		Placeholder - no idea how this is supposed to work, and no hardware to test.
	"""
	type = "LED"


#
# Sensor devices.
# This is complicated by iTachs reporting asynchronous changes differently
# from GC-100s. Yuck.
#
class Sensor(Device):
	""" A sensor input device. Can detect and signal low voltage events.

		GC-100, iTach, and iTach Flex handle SENSORNOTIFY each differently:
		- GC-100 sends "statechange" messages on the command channel.
		- iTach Sensor is statically configured for SENSOR or SENSORNOTIFY;
		  the latter sends UDP multicast packets(!) on fixed port UDP_PORT.
		- iTach Flex with relay/sensor pod cannot be statically configured to notify.
		  We must send a set_SENSORNOTIFY command and pick our own port. We always do,
		  and pick GC's "standard" port anyway.
	"""
	type = "SENSOR"
	_udp = None

	def __init__(self, netdev, address, specific):
		Device.__init__(self, netdev, address, specific)
		if specific == 'SENSOR_NOTIFY' and netdev.type == DEVICE_ITACH:
			# regular iTach sensor pod already configured to send UDP multicasts
			self._upd_register()
		elif specific == 'SENSOR' and netdev.type == DEVICE_FLEX:
			# relay/sensor pod. Need to configure for notifications
			self._udp_register()
			# request the common GC UDP port, no timeouts
			self.command("set_SENSORNOTIFY", str(UDP_PORT), "0")
	
	def _udp_register(self):
		# iTach SENSOR_NOTIFY uses a multicast UDP port *only* to notify (bad call)
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
		s.bind(('', UDP_PORT))
		# join multicast group
		mreq = struct.pack('4sl', socket.inet_aton(UDP_ADDRESS), socket.INADDR_ANY)
		s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
		self._udp = self.hostdev.control.datagram(s, callout=self._udp_calldown)

	def close(self):
		if self._udp:
			self._udp.close()

	def poll(self):
		""" Poll the sensor state. Will call back with the answer. """
		self.command("getstate")

	def _calldown(self, ctx, cmd, args):
		if cmd == "state" or cmd == "statechange":
			# polled or GC-100 style notify
			self.callout('state', args == "1")
		elif cmd == "SENSORNOTIFY":
			# reply to our (iTach Flex) notification enablement; just ignore it
			pass
		else:
			Device._calldown(self, ctx, cmd, args)

	def _udp_calldown(self, ctx, dgram = None):
		if ctx.error:
			self.callout(ctx)
		if ctx.state == 'DGRAM':
			if ctx.source[0] == self.hostdev.ipaddr:	# from our hostdev
				self.hostdev.idle_activity()			# (we got activity there)
				try:
					cmd, addr, value = dgram.strip().split(',')
					if cmd == "sensornotify":
						if addr == self.address:
							self.callout('state', value == "1")
				except Exception as e:
					# probably corrupted UDP garbage
					print("UDP message error (%s): %s" % (e, dgram))


#
# (Dry) Relay devices
#
class Relay(Device):
	""" A relay (dry contact) output device. There is no way to read its current state. """
	type = "RELAY"

	def poll(self):
		""" Poll the sensor state. Will call back with the answer. """
		self.command("getstate")

	def engage(self, active):
		""" Engage (True) or disengage (False) the relay. """
		self.command("setstate", "1" if active else "0")

	def _calldown(self, ctx, cmd, args):
		if cmd in ['state', 'statechange', 'setstate']:
			# polled or GC-100 style notify or cmd feedback
			self.callout('state', args == "1")
		else:
			Device._calldown(self, ctx, cmd, args)


#
# Serial devices.
# Note that this represents the meta-side of the connection; all the data flows
# through the separate TCP connection to port 4999+.
#
class Serial(Device):
	""" A serial port device. Provides transparent RS-232 bridging. """
	type = "SERIAL"

	def __init__(self, netdev, address, specific):
		Device.__init__(self, netdev, address, specific)
		self.data_port = netdev._serial_port
		netdev._serial_port += 1

	def connect(self, callout=None):
		""" Returns a TCPConnector for serial port access. """
		res = socket.getaddrinfo(self.hostdev.ipaddr, self.data_port, 0, socket.SOCK_STREAM)
		return self.hostdev.control.connector(res, callout=callout)

	def __repr__(self):
		return "<SERIAL %s/%s port %d>" % (self.name, self.specific, self.data_port)
