from __future__ import print_function
#
# IR Events as produced and understood by Global Cache devices
# (GC-IRL, GC-IRE, and iTach learning mode)
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ir
import re
import operator


class Error(ir.Error):
	""" GC device generated error. """


#
# GC-style Infrared signatures as produced by a receiver.
#
class Event(ir.Event):
	""" A GC-style IR event as produced by a receiver. """

	SCANRE=re.compile(r'(,\d+,\d+|[A-Z])')

	def __init__(self, text=None):
		""" create from GC-IRL text form. Incremental construction if text is None. """
		ir.Event.__init__(self)
		self._compdict = []						# up to 4, for A-D

		if text is not None:
			self.set_carrier(text.partition(',')[0])
			for item in self.SCANRE.findall(text):
				if item[0] == ',':
					self.add_pulse(item[1:])
				else:
					self.add_repeat(item)
			self.finish()

	def set_carrier(self, carrier):
		ir.Event.set_carrier(self, int(carrier))

	def add_pulse(self, pulse):					# "24,740"
		pulse = ir.Pulse(*map(int, pulse.split(',')))
		ir.Event.add_pulse(self, pulse)
		self._compdict.append(pulse)

	def add_repeat(self, code):
		""" Add a letter code. This is either a compression token or an error indicator. """
		if code == 'X': raise Error("buffer overflow")
		if code == 'Z': raise Error("aborted")
		index = ord(code) - ord('A')
		ir.Event.add_pulse(self, self._compdict[index])

	def finish(self):
		del self._compdict						# clean up
		ir.Event.finish(self)


#
# Basic regression test
#
if __name__ == "__main__":
	s1 = Event("36000,50,100,60,200BBABBAA,200,500B")
	s2 = Event("36000,51,97,60,199,59,200BABBAA,203,500B")
	s3 = Event("38000,50,100,60,200BBABBAA,200,500B")
	s4 = Event("36000,50,100,61,211BBABBAA,200,500B")
	s5 = Event("36000,50,100,60,200BBABAA,200,500B")
	assert(s1 == s1)
	assert(not (s1 != s1))
	assert(s1.ticks() == (50 + 100) * 4 + (60 + 200) * 6 + (200 + 500) - 200)
	assert(s1 == s2)
	assert(s1 != s3)
	assert(s1 != s4)
	assert(s1 != s5)
	try:
		s6 = Event("38000,50,100,61,203BBAABBZ")
		assert(False)
	except Error:
		pass
	print('pass: gcir basics')
