#
# ir - generic IR type (unrecognized scheme / raw recording)
#
# The Generic coding type imposes no interpretation on an Event.
# It does check for obvious trailing repetitions and compacts them;
# but beware that zero or one repetitions can't be recognized as such,
# and comparisons must needs err on the side of flexibility.
#
# This is typically used as the final, catch-all coding type
# after all more specific types have refused to touch an Event.
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ir
import ir.hex
from functools import reduce

name = "generic"
exactMatch = False

FINAL_OFF = 0.05				# [s] imposed off on last pulse (non-repeat case)
REPEAT_MIN_OFF = 0.01			# [s] min. off length to count as repeat sequence


class Symbol(ir.Symbol):
	""" A generic IR Symbol. Simply a collection of carrier, base train, and optional repeat train. """

	def __init__(self, carrier, base, repeat):
		ir.Symbol.__init__(self, module)
		self.carrier = carrier
		self.base = base
		self.repeat = repeat

	def __repr__(self):
		if self.repeat:
			return "<Symbol/generic:%d %s...%s>" % (self.carrier, self.base, self.repeat)
		else:
			return "<Symbol/generic:%d %s>" % (self.carrier, self.base)

	def matches(self, code):
		""" Match a generic IR code. Simply compare base trains and try to accommodate repeat patterns. """
		if not isinstance(code, Code): return False
		if code.carrier != self.carrier: return False
		if code.repeat:
			if code.base != self.base: return False
			if code.repeat != self.repeat: return False
			return True
		else:
			if code.base == self.base: return True
			if self.repeat and code.base == self.base + self.repeat: return True
			return False

	def code(self, repeats=0):
		return Code(self.carrier, self.base, self.repeat, repeats)

	#
	# The generic string representation is pronto hex-format
	#
	def __str__(self):
		return "<%s:%s>" % (name, ir.hex.encode(self))

	@staticmethod
	def from_text(text):
		return ir.hex.decodeSymbol(text)


#
# Our codes simply record what we've found
#
class Code(ir.Code):

	def __init__(self, carrier, base, repeat, count):
		ir.Code.__init__(self, module)
		self.carrier = carrier
		self.base = base
		self.repeat = repeat
		self.count = count

	def __str__(self):
		if self.repeat:
			return "<Code/generic:%d %s...%s{%d}>" % (self.carrier, self.base, self.repeat, self.count)
		else:
			return "<Code/generic:%d %s>" % (self.carrier, self.base)

	def __repr__(self): return str(self)

	def symbol(self):
		return Symbol(self.carrier, self.base, self.repeat)

	def form(self):
		return (self.base, self.repeat, self.count)


def claim(event):
	""" Claim any IR event as ours.

		GenericType accepts any IR event as type-compliant.
		It simply records the information contained and cleans it up a bit.
	"""
	train = event.pulsetrain

	# collect pulses into bins
	bins = []			# fuzzy collector bins
	index = []			# pattern to bin index
	final = train.pop() # keep it; it's special
	for pulse in train:
		for p in bins:
			if pulse == p['pulse']:
				p['count'] += 1
				p['on'] += pulse.on
				p['off'] += pulse.off
				break
		else:
			p = {'pulse': pulse, 'count': 1, 'on': pulse.on, 'off': pulse.off}
			bins.append(p)
		index.append(p)

	# average out each bin
	def smooth(sum, count): return int(round(float(sum) / count))
	for bin in bins:
		bin['pulse'] = ir.Pulse(smooth(bin['on'], bin['count']),
			smooth(bin['off'], bin['count']))

	# Regenerate pattern with bin averages. This also unifies Pulses (except the final one)
	train = [bin['pulse'] for bin in index] + [final]

	# look for a trailing repeat pattern
	plen = len(train)
	for rlen in range(2, plen/2):	# min 2; at least one repetition
		repeat = train[-rlen:]		# candidate repeat
		rpos = plen - 2 * rlen		# [...] ^ [rep-1] [rep]
		if train[rpos:rpos+rlen] != repeat:	# nothing repeated
			continue
		repeat[-1] = train[-rlen-1] # instantiate wildcard Pulse for this match
		if reduce(max, map(lambda x: x.off, repeat)) < REPEAT_MIN_OFF * event.carrier:
			# too short, probably just a trailing run of tokens
			continue
		# we have found a viable repeat - now extend it backwards as far as it will go
		while rpos > rlen and train[rpos-rlen:rpos] == repeat:
			rpos -= rlen
		return Code(event.carrier, train[0:rpos], repeat, (plen - rpos) / rlen)
	else:	# no repeat found
		train[-1] = train[-1].edit(int(FINAL_OFF * event.carrier))
		return Code(event.carrier, train, None, None)


#
# Fetch our own module reference
#
import sys
module = sys.modules[__name__]
