#
# ir_nec - NEC style CIR signals
#
# For better or worse, this coding type, invented by NEC,
# has been widely adopted by all comers, and we shouldn't blame
# NEC for the various perversions inflicted on it over the years.
# At heart, we're just requiring a 38kHz carrier and pulse trains
# that reduce to two known Pulse types arrayed in specific bit
# patterns.
#
# Copyright 2010-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import ir
from functools import reduce

name = "NEC"
exactMatch = True


#
# Characteristics of the NEC "protocol"
#
# These parameters are a compromise between "official" tightness
# (rejecting variations from competitors), and excessive looseness
# (matching everyone but losing definition). It's a "works for me"
# kind of compromise, to be sure.
#
CARRIER = 38000
LEADER = ir.Pulse(342,170)
ZERO = ir.Pulse(22,21)
ONE = ir.Pulse(22,63)
TRAILER = ir.Pulse(23,2200)
TRAILER_SLOP = 1600 # different hardware is VERY sloppy about TRAILER.off
REPEAT = [ir.Pulse(342,85), ir.Pulse(22,3640)] # does not vary


class Symbol(ir.Symbol):

	carrier = CARRIER

	def __init__(self, dev, func):
		ir.Symbol.__init__(self, module)
		self.device = dev
		self.function = func

	def matches(self, code):
		return (isinstance(code, Code) and
			self.device == code.device and
			self.function == code.function)

	def code(self, repeats=0):
		return Code(self.device, self.function, repeats)

	#
	# Canonical string form
	#
	def __str__(self):
		return "<%s:%04x/%02x>" % (name, self.device, self.function)

	def __repr__(self): return str(self)

	@staticmethod
	def from_text(text):
		dev, fun = text.split('/')
		return Symbol(int(dev, 16), int(fun, 16))


class Code(ir.Code):

	carrier = CARRIER

	def __init__(self, dev, func, repeat):
		ir.Code.__init__(self, module)
		self.device = dev
		self.function = func
		self.repeat = repeat

	def __str__(self):
		return "<Code:NEC %04x/%02x{%d}>" % (self.device, self.function, self.repeat)

	def __repr__(self): return str(self)

	def symbol(self):
		return Symbol(self.device, self.function)

	def form(self):
		""" Make abstract IR form. Returns (base, repeat, count) tuple. """
		func = self.function
		if (func & ~0xff) == 0:	# good parity
			func |= ((0xff ^ func) << 8)
		def bits(mask, count):
			for n in range(0, count): yield mask & 1; mask = mask >> 1
		base = ([LEADER]
		+	[ONE if x else ZERO for x in bits(self.device, 16)]
		+	[ONE if x else ZERO for x in bits(func, 16)]
		+	[TRAILER])
		if self.repeat:
			return (base, REPEAT, self.repeat)
		else:
			return (base, None, 0)


#
# Type check and encode
#
def claim(event):
	""" If event looks like an NEC signal, encode it as our Code. Return None if it doesn't. """

	#
	# An NEC pulsetrain looks like this:
	#	LEADER ( ZERO | ONE ){32} TRAILER ( REPEAT )*
	#
	try:
		if event.carrier != CARRIER: return None
		if event[0] != LEADER: return None
		if not TRAILER.__eq__(event[33], TRAILER_SLOP): return None

		pulses = event.pulsetrain

		# count and check repeat patterns
		repeats = (len(pulses) - 34) / len(REPEAT)
		if pulses[34:] != REPEAT * repeats: return None	# check repetitions

		# extract bits
		def bit(p):
			if p == ZERO: return 0
			elif p == ONE: return 1
			else: raise ValueError("invalid bit")
		code = map(bit, pulses[1:33])
		code.reverse() # low-endian

		# Code is now: [func:16][device:16]
		def bits(slice): return reduce(lambda a, b: a*2+b, slice, 0)
		func_check = bits(code[0:8])
		func = bits(code[8:16])
		dev = bits(code[16:32])
		if (func_check ^ func) != 0xff:
			func |= (func_check << 8)
		return Code(dev, func, repeats)

	except (ValueError, IndexError): return None


#
# Fetch our own module reference
#
import sys
module = sys.modules[__name__]
